package de.fhhannover.inform.trust.ifmapj_examples;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.Semaphore;

import org.w3c.dom.Document;

import de.fhhannover.inform.trust.ifmapj.IfmapJ;
import de.fhhannover.inform.trust.ifmapj.IfmapJHelper;
import de.fhhannover.inform.trust.ifmapj.binding.IfmapStrings;
import de.fhhannover.inform.trust.ifmapj.channel.ARC;
import de.fhhannover.inform.trust.ifmapj.channel.SSRC;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.identifier.Identifier;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentifierFactory;
import de.fhhannover.inform.trust.ifmapj.identifier.IpAddress;
import de.fhhannover.inform.trust.ifmapj.identifier.MacAddress;
import de.fhhannover.inform.trust.ifmapj.messages.PollResult;
import de.fhhannover.inform.trust.ifmapj.messages.PublishDelete;
import de.fhhannover.inform.trust.ifmapj.messages.RequestFactory;
import de.fhhannover.inform.trust.ifmapj.messages.ResultItem;
import de.fhhannover.inform.trust.ifmapj.messages.SearchResult;
import de.fhhannover.inform.trust.ifmapj.metadata.LocationInformation;
import de.fhhannover.inform.trust.ifmapj.metadata.StandardIfmapMetadataFactory;


/**
 * Test the behavior of a MAPS when there is a subscription that matches
 * a link which adds/removes a new subgraph to the poll result.
 * 
 */
public class PollWithSupgraphAddingRemovingExample {
	
	private IdentifierFactory ifac = IfmapJ.createIdentifierFactory();
	private RequestFactory rf = IfmapJ.createRequestFactory();
	private StandardIfmapMetadataFactory mf = IfmapJ.createStandardMetadataFactory();
	private SSRC ssrc;
	private ARC arc;
	private Semaphore mainSem, pollSem;
	
	private IpAddress ip;
	private MacAddress mac;
	Document ipMac;
	Document location;
	
	public PollWithSupgraphAddingRemovingExample() {
		// prepare identifiers
		ip = ifac.createIp4("192.168.0.1");
		mac = ifac.createMac("aa:bb:cc:dd:ee:ff");
		// prepare metadata
		ipMac = mf.createIpMac("2011-01-26T19:32:52Z", "2011-01-27T19:32:52Z", "trust-dhcp-server");
		List<LocationInformation> l = new ArrayList<LocationInformation>();
		LocationInformation li1 = new LocationInformation("GPS", "N 52.517526 E 13.44");
		LocationInformation li2 = new LocationInformation("ROOM", "326");
		l.add(li1); l.add(li2);
		location = mf.createLocation(l, "2011-01-27T19:32:52Z", "my location sensor");
	}
	
	private class Polling implements Runnable {

		@Override
		public void run() {
			try {
				PollResult pollResult;
			
				// main has to allow us to poll
				pollSem.acquire();
			
				// the searchResult
				System.out.print("[poll-thread] Polling for searchResult ... ");
				pollResult = arc.poll();
				if (pollResult.getSearchResults().size() < 1)
					System.out.println("Unexpected PollResult");
				else
					System.out.println("OK");
				IfmapJExamples.sleepSomeTime();
	
				// let the main thread continue
				mainSem.release();
				
				// wait for the main thread to allow us to continue
				pollSem.acquire();
				
				System.out.print("[poll-thread] Polling for updateResult ... ");
				pollResult = arc.poll();
				
				if (pollResult.getUpdateResults().size() < 1)
					System.out.println("Unexpected PollResult");
				else
					checkContainsExpectedMetadata(pollResult.getUpdateResults());
				IfmapJExamples.sleepSomeTime();
			
				// let the main-thread delete the metadata
				mainSem.release();
			
				// wait until stuff is deleted
				pollSem.acquire();
				
				System.out.print("[poll-thread] Polling for deleteResult ... ");
				pollResult = arc.poll();
				
				if (pollResult.getDeleteResults().size() < 1)
					System.out.println("Unexpected PollResult");
				else
					checkContainsExpectedMetadata(pollResult.getDeleteResults());
				IfmapJExamples.sleepSomeTime();
			
				// let the main thread end the session
				mainSem.release();
				
				IfmapJExamples.doLastPoll(arc);
				
			} catch (Exception e) {
				System.err.println("[poll-thread] Exception: " + e.getMessage());
				e.printStackTrace();
			}	
		}

		private void checkContainsExpectedMetadata(Collection<SearchResult> results) {
			boolean ipMacOk = false;
			boolean locationOk = false;
			// Check whether we find ip-mac link and mac identifier,
			// both containing a single metadata element
			for (SearchResult sr : results) {
				for (ResultItem ri : sr.getResultItems()) {
					Identifier idents[] = ri.getIdentifier();
					if (idents[1] == null) {
						if (idents[0] instanceof MacAddress && ri.getMetadata().size() == 1) {
							MacAddress mac = (MacAddress)idents[0];
							Document md = ri.getMetadata().iterator().next();
							locationOk = mac.getValue().equals("aa:bb:cc:dd:ee:ff")
							&& md.getFirstChild().getLocalName().equals("location");
						}
					} else {
						MacAddress mac = null;
						IpAddress ip = null;
						if (idents[0] instanceof IpAddress && idents[1] instanceof MacAddress) {
							ip = (IpAddress) idents[0];
							mac = (MacAddress) idents[1];
						} else if (idents[0] instanceof MacAddress && idents[1] instanceof IpAddress) {
							mac = (MacAddress) idents[0];
							ip = (IpAddress) idents[1];
						}
						
						if (mac != null && ip != null && ri.getMetadata().size() == 1) {
							Document md = ri.getMetadata().iterator().next();
							ipMacOk = md.getFirstChild().getLocalName().equals("ip-mac")
							&& ip.getValue().equals("192.168.0.1")
							&& mac.getValue().equals("aa:bb:cc:dd:ee:ff");
						}
					}
				}
			}
			
			if (ipMacOk && locationOk) {
				System.out.println("OK");
			} else {
				System.out.println("Could NOT find expected Metadata");
			}
		}
	}
	

	public void run (String args[]) throws InitializationException {
		
		System.out.println("====== RUNNING POLL WITH SUBGRAPH OPS EXAMPLE ======");
		
		mainSem = new Semaphore(0);
		pollSem = new Semaphore(0);

		// Use basic authentication
		ssrc = IfmapJ.createSSRC(
				Config.BASIC_AUTH_SERVER_URL,
				Config.BASIC_AUTH_USER,
				Config.BASIC_AUTH_PASSWORD,
				IfmapJHelper.getTrustManagers(
				getClass().getResourceAsStream(Config.TRUST_STORE_PATH),
				Config.TRUST_STORE_PASSWORD));
		
		arc = ssrc.getArc();
		
		try {
	
			System.out.print("[main-thread] newSession ... ");
			ssrc.newSession();
			System.out.println("OK");
			IfmapJExamples.sleepSomeTime();
		
			// Create a subscription beginning from 192.168.0.1
			System.out.print("[main-thread] subscribe ... ");
			ssrc.subscribe(rf.createSubscribeReq(rf.createSubscribeUpdate(
					"subscription", // name
					null,			// match-links
					10,				// maxDepth
					null,			// terminal-identifiers,
					null,			// max-size
					null,			// result-filter
					ip)));			// start identifier
			System.out.println("OK");
			IfmapJExamples.sleepSomeTime();
			
			// start the polling thread
			new Thread(new Polling()).start();
		
			// let it go
			pollSem.release();
	
			// continue if the poll thread has done its first polling
			mainSem.acquire();
			
			System.out.print("[main-thread] publishing ip-mac... ");
			ssrc.publish(rf.createPublishReq(rf.createPublishUpdate(
					ip, mac, ipMac)));
			System.out.println("OK");
			
			IfmapJExamples.sleepSomeTime();
			System.out.print("[main-thread] publishing location to mac... ");
			ssrc.publish(rf.createPublishReq(rf.createPublishUpdate(
					mac, location)));
			System.out.println("OK");
			IfmapJExamples.sleepSomeTime();
		
			
			// let the poll-thread do it's polling
			pollSem.release();
		
			// wait until the poll-thread is done
			mainSem.acquire();

			// delete ip-mac
			PublishDelete delete = rf.createPublishDelete(ip, mac, "meta:ip-mac");
			delete.addNamespaceDeclaration(IfmapStrings.STD_METADATA_PREFIX,
					IfmapStrings.STD_METADATA_NS_URI);
			
			System.out.print("[main-thread] deleting ... ");
			// delete the metadata
			ssrc.publish(rf.createPublishReq(delete));
			System.out.println("OK");
			IfmapJExamples.sleepSomeTime();
		
			// let the poll-thread get the deleteResult
			pollSem.release();
		
			// wait until it's done
			mainSem.acquire();

			IfmapJExamples.doLastEndSession("[main-thread]", ssrc);
			IfmapJExamples.sleepSomeTime();
		} catch (IfmapException e ) {
			System.err.println(e.getMessage());
		} catch (IfmapErrorResult e) {
			System.err.println(e);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
